package PureMessage::Service::Sendmail;
use File::stat;
use Socket;
use PureMessage::Version qw($PREFIX);
use ActiveState::Unix::ProcInfo qw(proc_info);
use PureMessage::Manager::Lang;
use PureMessage::Service::Config qw(is_stale);

use vars qw(@EXPORT @EXPORT_OK %EXPORT_TAGS);
use base 'Exporter';
our @EXPORT_OK = qw (
            getSendmailPID
            expect_sendmail_status
            get_sendmail_status
            refresh_mc
            refresh_mailertable
            save_mailer_table
            refresh_accesstable
            save_access_table
            save_milter_data
            smconfNewer
            isSendmailRunning
            isGateway
            makeGateway
            makeHub
            isMilterResponding
            hasAllPerms
            canStartStop
            canEditMilters
            canEditMailerTable
            canEditAccessTable
            checkSuidPerms
            new
            mc_line
            _timeout_parse
            _timeout_fmt
            run_suid
            run_edit_milter
            raise_suid_error
            edit_mailer_table
            add_mailer_table
            remove_mailer_table
            edit_access_table
            add_access_table
            remove_access_table
            change_config
            $sm
            @milters
            @mc
            $mc_mod
            @relay_hosts
            @access_table
);
%EXPORT_TAGS = (all => \@EXPORT_OK);

our $sm; 
$sm = "$PREFIX/sendmail";

my $mc = "$sm/etc/mail/sendmail.mc";
my $cf = "$sm/etc/mail/sendmail.cf";
my $mc_hub = "$sm/etc/mail/example-hub.mc";
my $mc_gw = "$sm/etc/mail/example-gateway.mc";
my $mailertable = "$sm/etc/mail/mailertable";
my $accesstable = "$sm/etc/mail/access";
my $sm_pid_file = "$sm/var/run/sendmail.pid";
my $suid_path="$PREFIX/sendmail/sbin/";

$ENV{PATH} .= "$ENV{PATH}:/usr/ccs/bin"
    if $^O eq "solaris" && $ENV{PATH} !~ m,(^|:)/usr/ccs/bin($|:),;

my $m4 = $ENV{M4} || "m4";
my $make = $ENV{MAKE} || "make";
my $editor = $ENV{EDITOR} || "vi";

# globals that keep the state
our @mc;
our $mc_mod;     # is @mc modified since it was read from $mc
our @milters;
our @relay_hosts;
our @access_table;

# the caller expects sendmail's status to be a certain state,
# this attempts to match that state by testing sendmail's
# status multiple times with a wait in between. This gives sendmail
# time to shut down or start up before we report on its state
sub expect_sendmail_status{
	my $expect = $_[0] || '';
	my $info = get_sendmail_status();
	if($expect){
		if($expect eq 'started' && !$info->{'pid'}){
			# try 3 times
			my $ok = 0;
			for(1..3){
				sleep 1;
				$info = get_sendmail_status();
				if($info->{'pid'}){
					$ok = 1;
					last;
				}
			}
			$info = undef unless($ok);
		}elsif($expect eq 'stopped' && $info->{'pid'}){
			# try 3 times
			my $ok = 0;
			for(1..3){
				sleep 1;
				$info = get_sendmail_status();
				unless($info->{pid}){
					$ok = 1;
					last;
				}
			}
			$info = undef if($ok);			
		}
	}
	return $info;
}

sub get_sendmail_status {
	return undef unless hasAllPerms();
        
	my $sm_pid = getSendmailPID();
	t("looking for a sendmail with a pid of: $sm_pid");
    my $ps = "UNIX95=1 /bin/ps";
    my $ps_opts;
    my $ps_cmd;
    if ($^O eq "freebsd") {
        $ps_opts = "axc";
		$ps_cmd = "command";
    }
    elsif($^O eq "hpux" || $^O eq "linux") {
		$ps_opts = "e";
		$ps_cmd = "comm";
    }
    elsif($^O eq "solaris"){
		$ps_opts = "eL";
		$ps_cmd = "comm";
    }
    elsif($^O eq "aix"){
		$ps_opts = "e";
		$ps_cmd = "comm";
    }

    my %info; 
	my $cli = "$ps -${ps_opts}o pgid,pid,ppid,vsz,$ps_cmd";
	t("call ps to find sendmail process using the command $cli");
    if (open(PS, "$cli | ")) {
        while (<PS>) {
			next unless(/sendmail|mta/);
            my($pgid, $pid, $ppid, $vsz, $cmd) = split(' ', $_);
			next unless ($pid == $sm_pid);
			t("found sendmail: $_");
            $info{count}++;
            $info{pid} = $pgid unless $info{pid};
            $info{vsz} += $vsz;
        }
    }
	close PS;

    my $stat = stat($sm_pid_file);
    my $run_time = $stat->[9];
    my $is_stale_time = is_stale('smtp') || 0;

    if ($is_stale_time > $run_time) {
        $info{'config'} = 'stale';
    }
    else {
        $info{'config'} = 'current';
    }
    
    return \%info;
}

sub refresh_mc {
    my $file = shift || $mc;
    @milters = ();
    $mc_mod = $file ne $mc;

    open(MC, $file) || die "Can't open $file: $!";
    @mc = <MC>;
    close(MC);

    my $n = 0;
    for (@mc) {
        next unless /INPUT_MAIL_FILTER/;
        my $spec = MilterSpec->new($_);
        next unless $spec;
        $spec->{line} = $n;
        push(@milters, $spec);
    }
    continue {
        $n++;
    }

}

sub refresh_mailertable {
    my $file = shift || $mailertable;
    @relay_hosts = ();
	my $linec = 0;
	
    open(MT, $file) || die "Can't open $file: $!";
	while(<MT>){
		$linec++;
		s/^\s*//;
		s/\s*$//;
		next unless($_); # no blank lines
		next if(/^\s*#/); # no comment only lines
		push(@relay_hosts, [$_,$linec]);
	}
    close(MT);
		
}

sub save_mailer_table{
	my @new_data = @_; # array  of rows
	t('save_mailer_table()');
			
	eval{
		t('refresh the mailer table data');
		refresh_mailertable();
		t('ok');
		my $rc = 0;
		while(my $row = shift @relay_hosts){
			t("checking row:" . $row->[0]);
			# check for edit - if the user removes a line, all lines below it will be edited as well
			my $data = shift @new_data;
			t("check user input row: $data");
			$data =~s/^\s*//;
			$data =~s/\s*$//;
			unless ($data){
				t("row was deleted, removing from relay_hosts");
				unshift(@relay_hosts, $row);
				last;
			}
			if($data ne $row->[0]){
				t("Editing data: $row->[0], $data");
				$rc = edit_mailer_table($row, $data); # FIXME : error should be based on error code
				die "Failed to edit access table: $rc" unless $rc == 0;
				t("ok, rc = $rc");
			}
		}

		# if $new_data still has members, then the user added information, 
		if((scalar @new_data) > 0){
			foreach my $data(@new_data){
				t("Adding new data: $data");
				$rc = add_mailer_table($data); # FIXME : error should be based on error code
				die "Failed to add to access table: $rc" unless $rc == 0;
				t("OK, rc = $rc");
			}
		}elsif((scalar @relay_hosts)>0){
			my @del_line = ();
			foreach my $data(@relay_hosts){
				push(@del_line, $data->[1]);
			}
			t("items to delete: @del_line");
			$rc = remove_mailer_table(@del_line); # FIXME : error should be based on error code
			die "Failed to remove from mailer table: $rc" unless $rc == 0;
			t("ok, rc = $rc");
			
		}
	};
	if($@){
		t("failed to save the mailer table: $@");
		die "Failed to seave the mailer table. $@";
	}
	t("save mailer table done ok");
	#webmin_log('edit', 'sendmail', 'mailer table', {'data'=>\@new_data});
	return 1;
}

sub refresh_accesstable{
    my $file = shift || $accesstable;
    @access_table = ();
	my $lcount = 0;
    open(AT, $file) || die "Can't open $file: $!";
	while(<AT>){
		$lcount++;
		s/^\s*//;
		s/\s*$//;
		next unless($_); # no blank lines
		next if(/^\s*#/); # no comment only lines
		next unless(/RELAY$/); # only take RELAY lines (no other option is supported by this editor at this time)
		/^(\S+)/; # take only the IP, host or domain name portion
		push(@access_table, [$1, $lcount]);
	}
    close(AT);
	
}
sub save_access_table{

	my @new_data = @_; # array  of rows

	eval{
		refresh_accesstable();
		while(my $row = shift @access_table){
			#print "Checking row: ", $row->[0]," ::\n";
			# check for edit - if the user removes a line, all lines below it will be edited as well
			my $data = shift @new_data;
			#print "Compare against data: $data\n";
			$data =~s/^\s*//;
			$data =~s/\s*$//;
			unless ($data){
				unshift(@access_table, $row); # put the data back (didn't use it)
				last;
			}
			if($data ne $row->[0]){
				#print "Editing data: ", $row->[0], ", $data ::\n";
				edit_access_table($row, $data); # FIXME : needs error handling
			}
		}

		# if $new_data still has members, then the user added information, 
		if((scalar @new_data) > 0){
			foreach my $data(@new_data){
				add_access_table($data); # FIXME : needs error handling
				#print "Adding new data: $data ::\n";
			}
		}elsif((scalar @access_table)>0){
			my @del_line = ();
			foreach my $data(@access_table){
				push(@del_line, $data->[1]);
			}
			# FIXME : needs error handling
			#print "items to delete: @del_line ::\n";
			remove_access_table(@del_line); # blindly removes all the remaining lines
			
		}
	};
	if($@){
		return 0;
		#	print "Error saving access table! $@\n";
	}
	#webmin_log('edit', 'sendmail', 'access table', {'data'=>\@new_data});
	return 1;
}


# calls the external programs that save the milter information
#
sub save_milter_data{
	my $action = shift;
	my $milter = shift;
	my $cli = '';
	
	if($action eq 'edit'){
		$cli = "-m '$milter->{'name'}' ";
		$cli .= "-s '$milter->{'socket'}' ";
		$cli .= "-t '$milter->{'enabled'}' ";
		$cli .= "-f $milter->{'fail'} ";
		$cli .= "-C '$milter->{'timeout'}{'C'}' ";
		$cli .= "-R '$milter->{'timeout'}{'R'}' ";
		$cli .= "-E '$milter->{'timeout'}{'E'}' ";
		$cli .= "-S '$milter->{'timeout'}{'S'}' ";
	}
	if($action eq 'add'){
		$cli = '-a ';
		$cli .= "-m '$milter->{'name'}' ";
		$cli .= "-s '$milter->{'socket'}' ";
		$cli .= "-t '$milter->{'enabled'}' ";		
		$cli .= "-f $milter->{'fail'} ";
		$cli .= "-C '$milter->{'timeout'}{'C'}' ";
		$cli .= "-R '$milter->{'timeout'}{'R'}' ";
		$cli .= "-E '$milter->{'timeout'}{'E'}' ";
		$cli .= "-S '$milter->{'timeout'}{'S'}' ";
	}
	my $rc = run_edit_milter($cli);
	#webmin_log($action, 'sendmail milter', $milter->{'name'}, $milter);
	return $rc;
}

# checks to see if the running senmail needs to be restarted because the config file
# has changed. This is not an issue if sendmail is not running (the next time it starts,
# the new settings will be applied).
sub smconfNewer{

	if(!isSendmailRunning()){
		return 0; # Not running, don't worry about it.
	}
	use File::stat;
	my $stconf = stat($mc);
	my $mtconf = stat($mailertable);
	my $acconf = stat($accesstable);
	my $stpid = stat($sm_pid_file);
	unless(defined $stpid) {return 0;} # sendmail hasn't been run yet
	if($stconf->mtime > $stpid->mtime || $mtconf->mtime > $stpid->mtime || $acconf->mtime > $stpid->mtime){
		return 1;
	}

	return 0;
}

sub isSendmailRunning{

	my $pid = getSendmailPID();
	return 0 if(!$pid);
	return 0 unless (proc_info ('root_pid'=>$pid) );
	return 1;
}

# check to see if the current configuration is in gateway mode.
#
sub isGateway{
	unless (-e $mc){
		return 0; # file doesn't exist, so it can't be a gateway
	}
	open FH, "<$mc" or die "Failed to open the sendmail mc config file.";
	while(<FH>){
		if (/\QFEATURE(`mailertable\E/){
			return 1;
		}
	}
	return 0;
}

sub makeGateway{
	#webmin_log('change sendmail', 'config', 'to gateway');
	return change_config('gateway');
}
sub makeHub{
	#webmin_log('change sendmail', 'config', 'to hub');
	return change_config('hub');
}

# 'ping' a milter by connecting to it's socket. If the connection
# succeeds then return 1. return 0 if the connection fails.
#
# This supports both tcp (inet:) and unix (local:) sockets.
#
sub isMilterResponding{
	my $full_port = shift;
	my $rerr_msg = shift;
	
	my ($type, $location) = split ':', $full_port;
	my ($port, $host);
	if($type eq 'inet'){
		($port, $host) = split '@', $location;
		
		 unless ($port && $host){
			$$rerr_msg = text('invalid_host_port');
			return 0;
		 }
		unless ($port =~ /\d/){
			$$rerr_msg = $text{'port_not_integer'};
			return 0;
		}
		my $address = gethostbyname($host);
		unless ($address){
			$$rerr_msg = $text{'host_not_found'};
			return 0;
		}
		my $paddr = sockaddr_in($port, $address);
		my $rc = socket(SOCK, PF_INET, SOCK_STREAM, getprotobyname('tcp'));
		unless($rc){
			$$rerr_msg = text('socket_failed',$!);
			return 0;
		}
		if(!connect(SOCK, $paddr)){
			$$rerr_msg = text('connect_failed', $!);
			close SOCK;
			return 0;
		}
		close SOCK;
		return 1;

	}elsif($type eq 'local'){
		unless ( -S $location ) {
			$$rerr_msg = $text{'local_socket_not_found'};
			return 0;
        }
		unless (socket(SOCK, PF_UNIX, SOCK_STREAM, 0)){
			$$rerr_msg = text('local_socket_failed', $!);
			return 0;
		}
		unless(connect(SOCK, sockaddr_un($rendezvous))){
			$$rerr_msg = text('local_connect_failed', $!);
		}
		close SOCK;
		return 1;
	}else{
		$$rerr_msg = $text{'invalid_socket_type'};
		return 0;
	}
           
}

# returns true if all suid programs are setup 
sub hasAllPerms{
	return (canStartStop() && canEditMilters() && canEditMailerTable() && canEditAccessTable());
}
# returns true if stopping and starting sendmail is possible
sub canStartStop{
	return checkSuidPerms("$PREFIX/sendmail/sbin/sm_start");
}
# returns true if editing milters is possible
sub canEditMilters{
	return checkSuidPerms("$PREFIX/sendmail/sbin/sm_milter_edit");
}
# returns true if editing the mailer table is possible
sub canEditMailerTable{
	return checkSuidPerms("$PREFIX/sendmail/sbin/sm_mailer_table");
}
# returns true if editing the access table is possible
sub canEditAccessTable{
	return checkSuidPerms("$PREFIX/sendmail/sbin/sm_access_table");
}
# performs several checks on a file to see if it has been setuid root so it can be used.
sub checkSuidPerms{
	my $file = shift;
	return 0 unless -e $file;
	my $fstat = stat($file);
	return 0 unless $fstat;
	return 0 unless $fstat->uid == 0;
	return 0 unless -u $file;
	return 1;
}


# SUID functions

sub run_suid {
    my @cmd = @_;
    my $result = `$suid_path/sm_start @cmd 2>&1`;
        raise_suid_error($?, $result) if $?;
    return $result;
}

sub run_edit_milter {
	my @cmd = @_;
	my $result = `$suid_path/sm_milter_edit @cmd 2>&1`;
	my $rc;
	$rc = $? >> 8 if $?;
	t("Failed to edit the milter, Code: $?, Message: $result, Command: " , join(',',@cmd)) if $?;
	#raise_suid_error($?, "$result") if $?;
	return $rc;
}

sub getSendmailPID {
    my $result = `$suid_path/sm_start -p 2>&1`;
	chomp($result);
    t("getSendmailPID result: $result");
	my $rc;
	$rc = $? >> 8 if $?;
	return $result;
}

sub raise_suid_error {
    my $exit = shift;
    my $out = shift;
    my $exit_value = $exit >> 8;
    my $signal = $exit & 127;
    my $dumped_core = $exit & 128;
    
    # XXX provide callback for custom error handlers
    
    #$r->raise_error("Root execution failed: exit value: ".$exit_value."<pre>$out</pre>");
}

sub edit_mailer_table{
	my $relay_hosts = shift; # array ref. [0] is the data, [1] is the line number
	my $new_data = shift; # new data text, host and domain are separated by spaces
	my ($domain, $host) = split ' ',$new_data;
	my ($old_domain, $old_host) = split ' ', $relay_hosts->[0];
	my @cmd = ("-e", "-D '$old_domain'", "-d '$domain'", "-h '$host'", "-l $relay_hosts->[1]");
	my $result = `$suid_path/sm_mailer_table @cmd 2>&1`;
	my $rc = 0;
	$rc = $? >> 8 if $?;
	return $rc;
}


sub add_mailer_table{
	my $data = shift; # new data, host and domain are sep. by spaces
	t('add_mailer_table()',"adding: $data");
	my ($domain, $host) = split ' ',$data;
	my @cmd = ("-a", "-d $domain", "-h $host");
	my $result = `$suid_path/sm_mailer_table @cmd 2>&1`;
	t("result: $result");
	my $rc = 0;
	$rc = $? >> 8 if $?;
	t("rc = $rc, \$? = $?");
	return $rc;
}

sub remove_mailer_table{
	# each element is a line numberd to delete
	my @cmd = ("-r", @_);
	my $result = `$suid_path/sm_mailer_table @cmd 2>&1`;
	my $rc = 0;
	$rc = $? >> 8 if $?;
	return $rc;
}


############################################
# Access table functions
sub edit_access_table{
	my $access_hosts = shift; # array ref. [0] is the data, [1] is the line number
	my $new_host = shift; # new data text
	my $old_host =$access_hosts->[0];

	$new_host .= "\tRELAY\n";

	my @cmd = ("-e", "-H '$old_host'", "-h '$new_host'", "-l $access_hosts->[1]");
	my $result = `$suid_path/sm_access_table @cmd 2>&1`;
	my $rc = 0;
	$rc = $? >> 8 if $?;
	return $rc;
}

sub add_access_table{
	my $host = shift; # new data
	$host .= "\tRELAY\n";
	my @cmd = ("-a", "-h '$host'");
	my $result = `$suid_path/sm_access_table @cmd 2>&1`;
	my $rc = 0;
	$rc = $? >> 8 if $?;
	return $rc;
}

sub remove_access_table{
	# each element is a line numberd to delete
	my @cmd = ("-r", @_);
	my $result = `$suid_path/sm_access_table @cmd 2>&1`;
	my $rc = 0;
	$rc = $? >> 8 if $?;
	return $rc;
}


########################################################
# change the sendmail operating mode from/to gateway and
# hub
sub change_config{
	my $mode = shift;
	my @cmd;
	
	t("changing the config file");
	if($mode eq 'gateway'){
		t("to gateway");
		@cmd = ("-g");
	}else{
		t("to hub");
		@cmd = ("-h");
	}
	
	my $result = `$suid_path/sm_start @cmd 2>&1`;
	t("result = $result");
	my $rc = 0;
	$rc = $? >> 8 if $?;
	t("rc = $rc");
	return $rc;
}

sub t { 
    # print "@_\n" 
}


# this should be shared with pmx-sendmail-conf somehow
BEGIN {
	
package MilterSpec;

our @timeout_specs = (
    [C => "Milter connect",             "5m"],
    [S => "Send information to milter", "2m"],
    [R => "Read reply from the milter", "4m"],
    [E => "End of message",             "8m"],
);

sub new {
    my $class = shift;
    my $self = bless {}, $class;
    return $self unless @_;
    
    my $mc_line = shift;

    # $mc_line will be something like this:
    # INPUT_MAIL_FILTER(`test1',`S=inet:33666@localhost, F=T, T=C:10s;S:10s;R:10s;E:10s')
	# or for disabled milters, like this:
	# dnl INPUT_MAIL_FILTER(`test1',`S=inet:33666@localhost, F=T, T=C:10s;S:10s;R:10s;E:10s')
	if($mc_line =~ s/^\s*INPUT_MAIL_FILTER\(//){
		$self->{enabled} = 1;
	}elsif($mc_line =~ s/^dnl\s*INPUT_MAIL_FILTER\(//){
		$self->{enabled} = 0;
	}else{
		return 0;
	}
	
    #return unless $mc_line =~ s/^\s*INPUT_MAIL_FILTER\(//;
    return unless $mc_line =~ s/\)\s*$//;
    

    return unless $mc_line =~ s/^\s*\`([^\']+)\'\s*,\s*//;
    $self->{name} = $1;

    return unless $mc_line =~ s/^\`//;
    return unless $mc_line =~ s/\'$//;

    my @params = split(/\s*,\s*/, $mc_line);
    #$self->{params} = \@params;

    my @extra;
    for (@params) {
	if (s/^S=//) {
	    $self->{socket} = $_;
	}
	elsif (s/^F=//) {
	    $self->{fail} = $_;
	}
	elsif (s/^T=//) {
	    my %timeout;
	    for (split(/\s*;\s*/, $_)) {
		if (s/^([A-Z])://) {
		    $timeout{$1} = _timeout_parse($_);
		}
		else {
		    warn "Can't parse timeout spec $_";
		}
	    }
	    $self->{timeout} = \%timeout;
	}
	else {
	    push(@extra, $_);
	}
    }

    $self->{extra_param} = \@extra if @extra;

    $self;
}

sub mc_line {
    my $self = shift;
    my @line;
    push(@line, "INPUT_MAIL_FILTER(`$self->{name}',`");
    my @param;
    push(@param, "S=" . ($self->{socket} || 'inet:3366@localhost'));
    push(@param, "F=" . ($self->{fail} || "T"));
    if (my $t = $self->{timeout}) {
	push(@param, "T=" . join(";", map { "$_:" . _timeout_fmt($t->{$_}) }
				          sort keys %$t));
    }
    push(@param, @{$self->{extra_param}}) if $self->{extra_param};
    push(@line, join(", ", @param));
    push(@line, "')\n");
    join("", @line)
}

sub _timeout_parse {
    my $t = shift;
    #$t =~ s/s$//;
    #$t *= 60 if $t =~ s/m$//;
	$t =~ s/m$//;
    $t *= 60 if $t =~ s/s$//;
    
	
    $t;
}

sub _timeout_fmt {
    my $t = shift;
    return 0 unless $t;
    #return $t/60 . "m" if $t%60 == 0;
    #return $t . "s";
    return $t/60 . "s" if $t%60 == 0;
    return $t . "m";
	
} 

} # package


1;

