package PureMessage::Service::smtp;
use strict;
use PureMessage::Manager::Config qw( %web_config );
use PureMessage::Version qw ($PREFIX);
use PureMessage::Service::Sendmail qw(run_suid isGateway get_sendmail_status getSendmailPID hasAllPerms);
use ActiveState::Unix::ProcInfo qw(proc_info);
my $suid_path = $PREFIX."/sendmail/sbin";
use base 'PureMessage::Service';

if ( $@ ) { die "Error: $@ $!\n" }

sub Available {
    my $class = shift;
    my @avail;
    push(@avail, $class->new(name => "Sendmail"));
    return @avail;
}

sub _bg {
    my $pid = fork();
    die "Can't fork: $!" unless defined $pid;
    return $pid if $pid;

    open(STDIN, "</dev/null");
    open(STDOUT, ">/dev/null");
    open(STDERR, ">/dev/null");
    return 0;
}

sub start {
    my($self, %opts) = @_;
    if ($opts{bg}) {
        if (my $pid = _bg()) {
           return $pid;
        }
    }
    run_suid(" -s start");
    exit if $opts{bg};
}

sub stop {
    my($self, %opts) = @_;
    if ($opts{bg}) {
        if (my $pid = _bg()) {
            return $pid;
        }
    }
    run_suid(" -s stop");
    exit if $opts{bg};
}

sub restart {
    my($self, %opts) = @_;
    if ($opts{bg}) {
        if (my $pid = _bg()) {
            return $pid;
        }
    }
    run_suid(" -s stop");
    run_suid(" -s start");
    exit if $opts{bg};
}

sub status {
    my $self = shift;
    my $info = get_sendmail_status();
    if ( hasAllPerms() ) {
        $info->{ready} = $self->running(); # XXX - ping ?
        $info->{summary} = "Configured as Mail " . ( isGateway() ? "Gateway" : "Hub" ); 
        $info->{running} = $self->running();
    }
    else {
        $info->{read_only} = 1 unless hasAllPerms();
        $info->{ready} = undef;
        $info->{summary} = "Permission denied.";
        $info->{running} = undef;
    }
    $info;
}

sub running {
    my $self = shift;
    return undef unless hasAllPerms();
    my $pid = getSendmailPID();
    return 0 if(!$pid);
    my $p = proc_info ('root_pid'=>$pid);
    return 0 unless scalar keys %$p > 0;
    return 1;
}

1;
